/* Set.tlh
   Author: Theo Norvell
   Date: 1996 June 22
   Updated 2004 June 18

   This is an implementation of the Set ADT presented in class.
*/
#ifndef SET_H
#define SET_H

#include <new>
#include <cassert>

using namespace std ;

//  Set
////////////
//  A class for representing set valued variables.
//  Abstract Data:  S --  a set
template <class ElType>
class Set {
    public :
    // Exceptions
        class SpaceException {} ;

    //Constructors

        Set(void) ;
            // Construct an empty set
                //Post: S' = {}
                //May fail.

            Set(const Set &src) ;
            // Construct a copy of S
                //Post: S' = src.S
                //May fail.

        //Destructors
            ~Set(void) ;

    //Mutators

            const Set& operator=(const Set &src) ;
            //Assign to this set.
                // Post: S' = src.S and Result = a reference to src
                // May fail.

            void insert(const ElType &el) ;
            // Insert a member into this set.
                // Post:  S' = S U {el}
                // May fail.
        
            void insertNewMember(const ElType &el) ;
            // Insert a member into this set.
            // Pre: not (el in S)
                // Post:  S' = S U {el}
                // May fail.

            void remove(const ElType &el) ;
            // Remove a member from this set.
                //Post: S' = { x in S | x != el }
                //May not fail.

    //Accessors

        int size() const ;
        // Report the size of the set.
        // Post:  Result = |S|  and  S' = S

        bool contains(const ElType &el) const ;
            // Report whether or not something is in the set or not.
                // Post: Result = (el in S)  and  S'=S

        bool empty(void) const ;
            // Is the set empty or not?
        // Post: Result=(S={}) and S'=S

            ElType getAny(void) const ;
            // Return the value of any member in the set.
        // Pre: S != {}
        // Post: (Result in S) and S'=S

        bool operator== (const Set<ElType> &other) const ;
        // The equality operator
        // Post: Result=(S = other.S) and S'=S

        bool operator!= (const Set<ElType> &other) const ;
        // The equality operator
        // Post: Result=(S != other.S) and S'=S

    private :
        /* Representation:  Sets are represented by singly linked
           lists of nodes with one member of the set in each node.
           Members are listed in arbitrary order and with no duplicates
        */
        struct SetNode { ElType datum ; SetNode *flink ; } ;
        SetNode *head ;
        SetNode **find_link(const ElType &el) const ; 
} ;

////////////////////////////////////
//
// IMPLEMENTATION
//
////////////////////////////////////


/**************************************
 Private member functions
**************************************/

// find_link( el )
/* If el is in the set, returns a pointer to the link to the
node to that contains el. Otherwise, returns a pointer to the
last (null) link.
*/
template<class ElType>
Set<ElType>::SetNode ** Set <ElType> ::find_link (const ElType &el) const
{   
    SetNode **linkP = (SetNode **) & head ;
    while( *linkP != 0 && (*linkP)->datum != el ) {
            linkP = & ((*linkP)->flink) ; }
    return linkP ;
}

/*************************************
 Public member functions
*************************************/

// Constructor  (void)
template <class ElType>
Set<ElType>:: Set (void)
{
        head = 0 ;
}

// Copy Constructor
template <class ElType>
Set<ElType>:: Set( const Set<ElType> &S )
{
    head = 0 ;
    SetNode * p = S.head ;
    while( p != 0 ) {
            insert( p->datum ) ;
            p = p->flink ; }
}

// Destructor
template <class ElType>
Set<ElType>:: ~Set( void )
{
    while( head != 0 ) {
        SetNode *zombie = head ;
            head = head->flink ;
        delete zombie ; }
}

// Assignment operator
template <class ElType>
const Set<ElType> &  Set<ElType>::operator=(const Set<ElType> &S)
{
    // Clean up the old values
            while( head != 0 ) {
                SetNode *zombie = head ;
                head = head->flink ;
                delete zombie ; }
    // Copy in the new value
        SetNode * p = S.head ;
            while( p != 0 ) {
                insert( p->datum ) ;
                p = p->flink ; }
    return S ;
}

// insert
template <class ElType>
void Set<ElType>:: insert (const ElType &el)
{   
    SetNode **linkP = find_link( el ) ;

    if( *linkP != 0 ) { } // Already in do nothing.
    else { // Must be added.
            SetNode *temp = new(nothrow) SetNode ;
            if( temp == 0 ) { // Failed to find space
                //Indicate failure
                throw SpaceException() ; }
            else {
                //Fill in the new node
                    temp->datum = el ;
                    temp->flink = 0 ;
                    *linkP = temp ; } }
    assert( *find_link( el ) != 0 ) ; //Check part of postcondition
}

// insertNewMember
template <class ElType>
void Set<ElType>:: insertNewMember (const ElType &el)
{   
    SetNode *temp = new(nothrow) SetNode ;
    if( temp == 0 ) { // Failed to find space
        //Indicate failure
           throw SpaceException() ; }
    else {
        //Fill in the new node
            temp->datum = el ;
            temp->flink = head ; 
            head = temp ; }
    assert( *find_link( el ) != 0 ) ; //Check part of postcondition
}

// remove
template <class ElType>
void Set<ElType>:: remove (const ElType &el) 
{   
    SetNode **linkP = find_link( el ) ;
    if( *linkP != 0 ) { // el is in the set.
            //Remove the element
                SetNode *zombie = *linkP ;
                *linkP = (*linkP)->flink ;
                delete zombie ; }
    assert( *find_link( el ) == 0 ) ; // Check it is gone.
}

// size
template <class ElType>
int Set<ElType>:: size() const
{
    int count = 0 ;
    SetNode *p = head ;
    while( p != 0 ) {
        count += 1 ;
        p = p->flink ; }
    return count ;
}

// contains
template <class ElType>
bool Set<ElType>:: contains( const ElType &el) const
{       
        return *find_link(el) != 0 ;
}

// empty
template <class ElType>
bool Set<ElType>:: empty(void) const
{       
        return head == 0 ;
}

// getAny
template <class ElType>
ElType Set<ElType>:: getAny(void) const
{       assert( head != 0 ) ;
        return head->datum ;
}

// ==
template <class ElType>
bool Set<ElType>:: operator== (const Set<ElType> &S ) const
{
    if(size() == S.size() ) {   
        for( SetNode *p = head ; p != 0 ; p = p->flink ) {
            if( ! S.contains( p->datum ) ) {
                return false ; } }
        return true ; }
    else {
        return false ; }
}

// !=
template <class ElType>
bool Set<ElType>:: operator!= (const Set<ElType> &S ) const
{
    return ! (*this == S) ;
}

#endif
